#ifndef _gm_stbar_h_
#define _gm_stbar_h_

/* find out what architecture we are compiling for */

#include "gm_config.h"

/****************************************************************
 * Store synchronization barrier Macro
 ****************************************************************/

/* In some cases, we put the barrier code inside the following functions: */
GM_ENTRY_POINT extern void __gm_stbar (void);

/******************
 * lanai
 ******************/

#if defined lanai3 || defined lanai7 || defined lanai

 /* ": : :" for C++ compat. */
#define GM_STBAR() asm volatile ("! GM_STBAR" : : :"memory")
#define GM_READBAR() asm volatile ("! GM_READBAR" : : :"memory")
#define GM_WRITEBAR() asm volatile ("! GM_WRITEBAR" : : :"memory")

/****************
 * sparc
 ****************/

#elif GM_CPU_sparc || GM_CPU_sparc64
#ifdef __GNUC__
#define GM_STBAR( ) __gm_stbar ();
#define GM_READBAR() asm ("membar " "#LoadLoad | #LoadStore" : : : "memory")
#define GM_WRITEBAR() asm ("membar " "#StoreLoad | #StoreStore" : : : "memory")
#elif (defined (__SUNPRO_C) || defined (__SUNPRO_CC))
#define GM_STBAR( ) __gm_stbar ()
#define GM_READBAR() asm ("membar #LoadLoad | #LoadStore")
#define GM_WRITEBAR() asm ("membar #StoreLoad | #StoreStore")
#else
#error Do not know how to emit the "stbar" instruction under this compiler.
#endif

/****************
 * x86
 ****************/

#elif GM_CPU_x86
#if defined _MSC_VER
#define GM_STBAR( ) __gm_stbar ()
#define GM_READBAR() GM_STBAR()
#define GM_WRITEBAR() GM_STBAR()
#elif defined __GNUC__
#if GM_OS_SOLARIS
/* No obvious way to disable write-combining mapping on Solaris/i386,
 ensure proper semantics by making GM_STBAR much stronger */
#define GM_STBAR() asm ("lock; addl $0,0(%%esp)": : :"memory")
#else
#define GM_STBAR() asm ("": : :"memory")
#endif
#define GM_READBAR() asm ("": : :"memory")
#define GM_WRITEBAR() asm ("": : :"memory")
#elif defined __PGI
#define GM_STBAR()
#define GM_READBAR() GM_STBAR()
#define GM_WRITEBAR() GM_STBAR()
#elif defined __INTEL_COMPILER
#define GM_STBAR()
#define GM_READBAR() GM_STBAR()
#define GM_WRITEBAR() GM_STBAR()
#else
#define GM_STBAR()
#define GM_READBAR() GM_STBAR()
#define GM_WRITEBAR() GM_STBAR()
#endif

/****************
 * ia64
 ****************/

#elif GM_CPU_ia64
#if defined _MSC_VER
GM_ENTRY_POINT extern void __gm_stbar (void);
#define GM_STBAR( ) __gm_stbar ()
#define GM_READBAR() GM_STBAR()
#define GM_WRITEBAR() GM_STBAR()
#elif defined __GNUC__
#define GM_STBAR() asm volatile ("mf": : :"memory") /* ": : :" for C++ */
#define GM_READBAR() GM_STBAR()
#define GM_WRITEBAR() GM_STBAR()
#elif defined __INTEL_COMPILER
#define GM_STBAR( ) __gm_stbar ()
#define GM_READBAR() GM_STBAR()
#define GM_WRITEBAR() GM_STBAR()
#else
#error Do not know how to emit the "mf" instruction with this compiler.
#endif

/****************
 * alpha
 ****************/

#elif GM_CPU_alpha
#ifdef __GNUC__
#define GM_STBAR()  asm volatile ("mb": : :"memory") /* ": : :" for C++ */
#define GM_READBAR() asm volatile ("mb": : :"memory")
#define GM_WRITEBAR() asm volatile ("wmb": : :"memory")
#elif defined __DECC || defined __DECCXX
#if !GM_KERNEL
#include <c_asm.h>
#define GM_STBAR() asm ("mb")
#define GM_READBAR() asm ("mb")
#define GM_WRITEBAR() asm ("wmb")
#else
#include <sys/types.h>
#define GM_STBAR() mb()
#define GM_READBAR() mb()
#define GM_WRITEBAR() mb()
#endif
#else
#error Do not know how to emit the "mb" instruction with this compiler.
#endif

/****************
 * powerpc
 ****************/

#elif GM_CPU_powerpc
#ifdef __GNUC__
/* can't use -ansi for vxworks ccppc or this will fail with a syntax error */
#define GM_STBAR()  asm volatile ("sync": : :"memory") /* ": : :" for C++ */
#define GM_READBAR() asm volatile ("sync": : :"memory")
#define GM_WRITEBAR() asm volatile ("eieio": : :"memory")
#else
#if	GM_OS_AIX4
extern void __iospace_eieio(void); 
extern void __iospace_sync(void);  
#define GM_STBAR()   __iospace_sync ()
#define GM_READBAR() __iospace_sync ()
#define GM_WRITEBAR() __iospace_eieio ()
#else	/* GM_OS_AIX4 */
#error Do not know how to make a store barrier for this system
#endif	/* GM_OS_AIX4 */
#endif

/****************
 * mips
 ****************/

#elif GM_CPU_mips
#if GM_KERNEL
void flushbus(void);		/* hack to avoid including <sys/systm.h> */
#define GM_STBAR() flushbus();	/* kernel */
#define GM_READBAR() GM_STBAR()
#define GM_WRITEBAR() GM_STBAR()
#else
void gm_sync(void);		/* the myricom-provided flushbus() equiv. */
#define GM_STBAR() gm_sync();	/* user */
#define GM_READBAR() GM_STBAR()
#define GM_WRITEBAR() GM_STBAR()
#endif

/*****************
 * HP-PA RISC 
 *****************/
#elif GM_CPU_hppa
#define GM_STBAR()
#define GM_READBAR() GM_STBAR()
#define GM_WRITEBAR() GM_STBAR()

#endif /* various architectures */

#endif /* _gm_stbar_h_ */

